@@ -19,17 +19,21 @@ with keepassc.  If not, see <http://www.gnu.org/licenses/>.
 '''
 
 import curses as cur
+import logging
 from curses.ascii import NL, DEL, SP
 from datetime import date
 from os import chdir, getcwd, getenv, geteuid, makedirs, remove
-from os.path import expanduser, isfile, isdir, realpath
+from os.path import expanduser, isfile, isdir, realpath, join
 from pwd import getpwuid
 from random import sample
-from socket import gethostname
+from socket import gethostname, socket, AF_INET, SOCK_STREAM, SHUT_RDWR
 from sys import exit
 
-from kppy import KPDB, KPError
+from kppy.database import KPDBv1
+from kppy.exceptions import KPError
 
+from keepassc.conn import *
+from keepassc.client import Client
 from keepassc.editor import Editor
 from keepassc.helper import parse_config, write_config
 from keepassc.filebrowser import FileBrowser
@@ -50,24 +54,19 @@ class Control(object):
         try:
             self.config_home = realpath(expanduser(getenv('XDG_CONFIG_HOME')))
         except:
-            self.config_home = realpath(expanduser('~/.config/keepassc/'))
-        else:
-            if self.config_home[-1] == '/':
-                self.config_home += 'keepassc/config'
-            else:
-                self.config_home += '/keepassc/config'
+            self.config_home = realpath(expanduser('~/.config'))
+        finally:
+            self.config_home = join(self.config_home, 'keepassc', 'config')
 
         try:
             self.data_home = realpath(expanduser(getenv('XDG_DATA_HOME')))
         except:
-            self.data_home = realpath(expanduser('~/.local/share/keepassc/'))
-        else:
-            if self.data_home[-1] == '/':
-                self.data_home += 'keepassc/'
-            else:
-                self.data_home += '/keepassc/'
-        self.last_home = self.data_home + 'last'
-        self.key_home = self.data_home + 'key'
+            self.data_home = realpath(expanduser('~/.local/share/'))
+        finally:
+            self.data_home = join(self.data_home, 'keepassc')
+        self.last_home = join(self.data_home, 'last')
+        self.remote_home = join(self.data_home, 'remote')
+        self.key_home = join(self.data_home, 'key')
 
         self.config = parse_config(self)
 
@@ -98,21 +97,23 @@ class Control(object):
         cur.use_default_colors()
         cur.init_pair(1, -1, -1)
         cur.init_pair(2, 2, -1)
-        cur.init_pair(3, 0, 1)
+        cur.init_pair(3, -1, 1)
         cur.init_pair(4, 6, -1)
         cur.init_pair(5, 0, 6)
         cur.init_pair(6, 0, 7)
+        cur.init_pair(7, 1, -1)
         self.stdscr.bkgd(1)
         self.ysize, self.xsize = self.stdscr.getmaxyx()
 
         self.group_win = cur.newwin(self.ysize - 1, int(self.xsize / 3),
                                     1, 0)
-        self.entry_win = cur.newwin(int(2 * (self.ysize - 1) / 3),
+        # 11 is the y size of info_win
+        self.entry_win = cur.newwin((self.ysize - 1) - 11,
                                     int(2 * self.xsize / 3),
                                     1, int(self.xsize / 3))
-        self.info_win = cur.newwin(int((self.ysize - 1) / 3),
+        self.info_win = cur.newwin(11,
                                    int(2 * self.xsize / 3),
-                                   int(2 * (self.ysize - 1) / 3),
+                                   (self.ysize - 1) - 11,
                                    int(self.xsize / 3))
         self.group_win.keypad(1)
         self.entry_win.keypad(1)
@@ -126,12 +127,11 @@ class Control(object):
         self.ysize, self.xsize = self.stdscr.getmaxyx()
         self.group_win.resize(self.ysize - 1, int(self.xsize / 3))
         self.entry_win.resize(
-            int(2 * (self.ysize - 1) / 3), int(2 * self.xsize / 3))
-        self.info_win.resize(
-            int((self.ysize - 1) / 3), int(2 * self.xsize / 3))
+            self.ysize - 1 - 11, int(2 * self.xsize / 3))
+        self.info_win.resize(11, int(2 * self.xsize / 3))
         self.group_win.mvwin(1, 0)
         self.entry_win.mvwin(1, int(self.xsize / 3))
-        self.info_win.mvwin(int(2 * (self.ysize - 1) / 3), int(self.xsize / 3))
+        self.info_win.mvwin((self.ysize - 1) - 11, int(self.xsize / 3))
 
     def any_key(self):
         '''If any key is needed.'''
@@ -176,6 +176,60 @@ class Control(object):
         finally:
             self.stdscr.refresh()
 
+    def draw_help(self, *text):
+        """Draw a help
+
+        *text are arbitary string
+
+        """
+
+        if len(text) > self.ysize -1:
+            length = self.ysize - 1
+            offset = 0
+            spill = len(text) - self.ysize + 2
+        else:
+            length = len(text)
+            offset = 0
+            spill = 0
+
+        while True:
+            try:
+                self.draw_text(False)
+                for i in range(length):
+                    self.stdscr.addstr(
+                    i + 1, 0, text[(i + offset)])
+            except:
+                pass
+            finally:
+                self.stdscr.refresh()
+            try:
+                e = self.stdscr.getch()
+            except KeyboardInterrupt:
+                e = 4
+
+            if e == cur.KEY_DOWN:
+                if offset < spill:
+                    offset += 1
+            elif e == cur.KEY_UP:
+                if offset > 0:
+                    offset -= 1
+            elif e == NL:
+                return
+            elif e == cur.KEY_RESIZE:
+                self.resize_all()
+                if len(text) > self.ysize -1:
+                    length = self.ysize - 1
+                    offset = 0
+                    spill = len(text) - self.ysize + 2
+                else:
+                    length = len(text)
+                    offset = 0
+                    spill = 0
+            elif e == 4:
+                if self.db is not None:
+                    self.db.close()
+                self.close()
+
     def get_password(self, std, needed=True):
         '''This method is used to get a password.
 
@@ -192,6 +246,116 @@ class Control(object):
         else:
             return password
 
+    def get_authentication(self):
+        """Get authentication credentials"""
+
+        while True:
+            if (self.config['skip_menu'] is False or
+                (self.config['rem_db'] is False and
+                 self.config['rem_key'] is False)):
+                auth = self.gen_menu(1, (
+                                     (1, 0, 'Use a password (1)'),
+                                     (2, 0, 'Use a keyfile (2)'),
+                                     (3, 0, 'Use both (3)')),
+                                    (5, 0, 'Press \'F5\' to go back to main '
+                                           'menu'))
+            else:
+                self.draw_text(False)
+                auth = 3
+            if auth is False:
+                return False
+            elif auth == -1:
+                self.close()
+            if auth == 1 or auth == 3:
+                if self.config['skip_menu'] is True:
+                    needed = False
+                else:
+                    needed = True
+                password = self.get_password('Password: ', needed = needed)
+                if password is False:
+                    self.config['skip_menu'] = False
+                    continue
+                elif password == -1:
+                    self.close()
+                # happens only if self.config['skip_menu'] is True
+                elif password == "":
+                    password = None
+                if auth != 3:
+                    keyfile = None
+            if auth == 2 or auth == 3:
+                # Ugly construct but works
+                # "if keyfile is False" stuff is needed to implement the
+                # return to previous screen stuff
+                # Use similar constructs elsewhere
+                while True:
+                    self.get_last_key()
+                    if (self.last_key is None or
+                            self.config['rem_key'] is False):
+                        ask_for_lf = False
+                    else:
+                        ask_for_lf = True
+
+                    keyfile = FileBrowser(self, ask_for_lf, True, 
+                                          self.last_key)()
+                    if keyfile is False:
+                        break
+                    elif keyfile == -1:
+                        self.close()
+                    elif not isfile(keyfile):
+                        self.draw_text(False,
+                                       (1, 0, 'That\'s not a file'),
+                                       (3, 0, 'Press any key.'))
+                        if self.any_key() == -1:
+                            self.close()
+                        continue
+                    break
+                if keyfile is False:
+                    continue
+                if auth != 3:
+                    password = None
+                if self.config['rem_key'] is True:
+                    if not isdir(self.key_home[:-4]):
+                        if isfile(self.key_home[:-4]):
+                            remove(self.key_home[:-4])
+                        makedirs(self.key_home[:-4])
+                    handler = open(self.key_home, 'w')
+                    handler.write(keyfile)
+                    handler.close()
+            break
+        return (password, keyfile)
+
+    def get_last_db(self):
+        if isfile(self.last_home) and self.config['rem_db'] is False:
+            remove(self.last_home)
+            self.last_file = None
+        elif isfile(self.last_home):
+            try:
+                handler = open(self.last_home, 'r')
+            except Exception as err:
+                self.last_file = None
+                print(err.__str__())
+            else:
+                self.last_file = handler.readline()
+                handler.close()
+        else:
+            self.last_file = None
+
+    def get_last_key(self):
+        if isfile(self.key_home) and self.config['rem_key'] is False:
+            remove(self.key_home)
+            self.last_key = None
+        elif isfile(self.key_home):
+            try:
+                handler = open(self.key_home, 'r')
+            except Exception as err:
+                self.last_key = None
+                print(err.__str__())
+            else:
+                self.last_key = handler.readline()
+                handler.close()
+        else:
+            self.last_key = None
+
     def gen_pass(self):
         '''Method to generate a password'''
 
@@ -238,7 +402,7 @@ class Control(object):
                 edit = ''
                 e = cur.KEY_BACKSPACE
                 while e != NL:
-                    if e == cur.KEY_BACKSPACE and len(edit) != 0:
+                    if (e == cur.KEY_BACKSPACE or e == DEL) and len(edit) != 0:
                         edit = edit[:-1]
                     elif e == cur.KEY_BACKSPACE or e == DEL:
                         pass
@@ -279,7 +443,7 @@ class Control(object):
                 edit = ''
                 e = cur.KEY_BACKSPACE
                 while e != NL:
-                    if e == cur.KEY_BACKSPACE or e == DEL and len(edit) != 0:
+                    if (e == cur.KEY_BACKSPACE or e == DEL) and len(edit) != 0:
                         edit = edit[:-1]
                     elif e == cur.KEY_BACKSPACE or e == DEL:
                         pass
@@ -334,7 +498,7 @@ class Control(object):
             edit = ''
             e = cur.KEY_BACKSPACE
             while e != NL:
-                if e == cur.KEY_BACKSPACE or e == DEL and len(edit) != 0:
+                if (e == cur.KEY_BACKSPACE or e == DEL) and len(edit) != 0:
                     edit = edit[:-1]
                 elif e == cur.KEY_BACKSPACE or e == DEL:
                     pass
@@ -409,13 +573,13 @@ class Control(object):
             break
         return (y, mon, d)
 
-    def get_num(self, std='', edit=''):
+    def get_num(self, std='', edit='', length=4):
         '''Method to get a number'''
 
-        edit = ''
-        e = cur.KEY_BACKSPACE
+        edit = edit
+        e = 60 # just an unrecognized letter
         while e != NL:
-            if e == cur.KEY_BACKSPACE and len(edit) != 0:
+            if (e == cur.KEY_BACKSPACE or e == DEL) and len(edit) != 0:
                 edit = edit[:-1]
             elif e == cur.KEY_BACKSPACE or e == DEL:
                 pass
@@ -425,7 +589,7 @@ class Control(object):
                 self.resize_all()
             elif e == cur.KEY_F5:
                 return False
-            elif len(edit) < 4 and e >= 48 and e <= 57:
+            elif len(edit) < length and e >= 48 and e <= 57:
                 edit += chr(e)
             self.draw_text(False,
                            (1, 0, std + edit))
@@ -575,9 +739,10 @@ class Control(object):
                  (7, 0, 'Use directly password and key if one of the two '
                         'above is True: ' +
                   str(self.config['skip_menu'])),
-                 (8, 0, 'Generate default configuration'),
-                 (9, 0, 'Write config')),
-                (11, 0, 'Automatic locking works only for saved databases!'))
+                 (8, 0, 'Pin server certificate: ' + str(self.config['pin'])),
+                 (9, 0, 'Generate default configuration'),
+                 (10, 0, 'Write config')),
+                (12, 0, 'Automatic locking works only for saved databases!'))
             if menu == 1:
                 if self.config['del_clip'] is True:
                     self.config['del_clip'] = False
@@ -622,14 +787,20 @@ class Control(object):
                 elif self.config['skip_menu'] is False:
                     self.config['skip_menu'] = True
             elif menu == 8:
+                if self.config['pin'] is True:
+                    self.config['pin'] = False
+                elif self.config['pin'] is False:
+                    self.config['pin'] = True
+            elif menu == 9:
                 self.config = {'del_clip': True,  # standard config
                                'clip_delay': 20,
                                'lock_db': True,
                                'lock_delay': 60,
                                'rem_db': True,
                                'rem_key': False,
-                               'skip_menu': False}
-            elif menu == 9:
+                               'skip_menu': False,
+                               'pin': True}
+            elif menu == 10:
                 write_config(self, self.config)
                 return True
             elif menu is False:
@@ -664,62 +835,40 @@ class Control(object):
         finally:
             self.stdscr.refresh()
 
-    def get_last_db(self):
-        if isfile(self.last_home):
-            try:
-                handler = open(self.last_home, 'r')
-            except Exception as err:
-                self.last_file = None
-                print(err.__str__())
-            else:
-                self.last_file = handler.readline()
-                handler.close()
-        else:
-            self.last_file = None
-
-    def get_last_key(self):
-        if isfile(self.key_home):
-            try:
-                handler = open(self.key_home, 'r')
-            except Exception as err:
-                self.key_file = None
-                print(err.__str__())
-            else:
-                self.last_key = handler.readline()
-                handler.close()
-        else:
-            self.key_file = None
-
-    def main_loop(self, kdb_file=None):
+    def main_loop(self, kdb_file=None, remote = False):
         '''The main loop. The program alway return to this method.'''
 
-        # This is needed to remember last database and open it directly
-        self.get_last_db()
+        if remote is True:
+            self.remote_interface()
+        else:
+            # This is needed to remember last database and open it directly
+            self.get_last_db()
 
-        if kdb_file is not None:
-            self.cur_dir = kdb_file
-            if self.open_db(True) is True:
-                db = DBBrowser(self)
-                del db
-                last = self.cur_dir.split('/')[-1]
-                self.cur_dir = self.cur_dir[:-len(last) - 1]
-        elif self.last_file is not None and self.config['rem_db'] is True:
-            self.cur_dir = self.last_file
-            if self.open_db(True) is True:
-                db = DBBrowser(self)
-                del db
-                last = self.cur_dir.split('/')[-1]
-                self.cur_dir = self.cur_dir[:-len(last) - 1]
+            if kdb_file is not None:
+                self.cur_dir = kdb_file
+                if self.open_db(True) is True:
+                    db = DBBrowser(self)
+                    del db
+                    last = self.cur_dir.split('/')[-1]
+                    self.cur_dir = self.cur_dir[:-len(last) - 1]
+            elif self.last_file is not None and self.config['rem_db'] is True:
+                self.cur_dir = self.last_file
+                if self.open_db(True) is True:
+                    db = DBBrowser(self)
+                    del db
+                    last = self.cur_dir.split('/')[-1]
+                    self.cur_dir = self.cur_dir[:-len(last) - 1]
 
         while True:
             self.get_last_db()
             menu = self.gen_menu(1, ((1, 0, 'Open existing database (1)'),
                                   (2, 0, 'Create new database (2)'),
-                                  (3, 0, 'Configuration (3)'),
-                                  (4, 0, 'Quit (4)')),
-                                 (6, 0, 'Type \'F1\' for help inside the file '
+                                  (3, 0, 'Connect to a remote database(3)'),
+                                  (4, 0, 'Configuration (4)'),
+                                  (5, 0, 'Quit (5)')),
+                                 (7, 0, 'Type \'F1\' for help inside the file '
                                         'or database browser.'),
-                                 (7, 0, 'Type \'F5\' to return to the previous'
+                                 (8, 0, 'Type \'F5\' to return to the previous'
                                         ' dialog at any time.'))
             if menu == 1:
                 if self.open_db() is False:
@@ -734,7 +883,10 @@ class Control(object):
                                          (1, 0, 'Use a password (1)'),
                                          (2, 0, 'Use a keyfile (2)'),
                                          (3, 0, 'Use both (3)')))
-                    self.db = KPDB(new=True)
+                    password = None
+                    confirm = None
+                    filepath = None
+                    self.db = KPDBv1(new=True)
                     if auth is False:
                         break
                     elif auth == -1:
@@ -800,8 +952,10 @@ class Control(object):
                         self.db = None
                     break
             elif menu == 3:
+                self.remote_interface()
+            elif menu == 4:
                 self.gen_config_menu()
-            elif menu == 4 or menu is False or menu == -1:
+            elif menu == 5 or menu is False or menu == -1:
                 self.close()
 
     def open_db(self, skip_fb=False):
@@ -816,72 +970,11 @@ class Control(object):
             else:
                 self.cur_dir = filepath
 
-        while True:
-            if (self.config['skip_menu'] is False or 
-                (self.config['rem_db'] is False and
-                 self.config['rem_key'] is False)):
-                auth = self.gen_menu(1, (
-                                     (1, 0, 'Use a password (1)'),
-                                     (2, 0, 'Use a keyfile (2)'),
-                                     (3, 0, 'Use both (3)')),
-                                    (5, 0, 'Press \'F5\' to go back to main '
-                                           'menu'))
-            else:
-                self.draw_text(False)
-                auth = 3
-            if auth is False:
-                return False
-            elif auth == -1:
-                self.close()
-            if auth == 1 or auth == 3:
-                password = self.get_password('Password: ')
-                if password is False:
-                    self.config['skip_menu'] = False
-                    continue
-                elif password == -1:
-                    self.close()
-                if auth != 3:
-                    keyfile = None
-            if auth == 2 or auth == 3:
-                # Ugly construct but works
-                # "if keyfile is False" stuff is needed to implement the
-                # return to previous screen stuff
-                # Use similar constructs elsewhere
-                while True:
-                    if self.config['rem_key'] is True:
-                        self.get_last_key()
-                    if (self.last_key is None or
-                            self.config['rem_key'] is False):
-                        ask_for_lf = False
-                    else:
-                        ask_for_lf = True
+        ret = self.get_authentication()
+        if ret is False:
+            return False
+        password, keyfile = ret
 
-                    keyfile = FileBrowser(self, ask_for_lf, True, self.last_key)()
-                    if keyfile is False:
-                        break
-                    elif keyfile == -1:
-                        self.close()
-                    elif not isfile(keyfile):
-                        self.draw_text(False,
-                                       (1, 0, 'That\'s not a file'),
-                                       (3, 0, 'Press any key.'))
-                        if self.any_key() == -1:
-                            self.close()
-                        continue
-                    break
-                if keyfile is False:
-                    continue
-                if auth != 3:
-                    password = None
-                if self.config['rem_key'] is True:
-                    if not isdir(self.key_home[:-4]):
-                        if isfile(self.key_home[:-4]):
-                            remove(self.key_home[:-4])
-                        makedirs(self.key_home[:-4])
-                    handler = open(self.key_home, 'w')
-                    handler.write(keyfile)
-                    handler.close()
-            break
         try:
             if isfile(self.cur_dir + '.lock'):
                 self.draw_text(False,
@@ -908,7 +1001,8 @@ class Control(object):
                         break
             else:
                 read_only = False
-            self.db = KPDB(self.cur_dir, password, keyfile, read_only)
+            self.db = KPDBv1(self.cur_dir, password, keyfile, read_only)
+            self.db.load()
             return True
         except KPError as err:
             self.draw_text(False,
@@ -920,74 +1014,243 @@ class Control(object):
             self.cur_dir = self.cur_dir[:-len(last) - 1]
             return False
 
+    def remote_interface(self, ask_for_agent = True, agent = False):
+        if ask_for_agent is True and agent is False:
+            use_agent = self.gen_menu(1, ((1, 0, 'Use agent (1)'),
+                                          (2, 0, 'Use no agent (2)')))
+        elif agent is True:
+            use_agent = 1
+        else:
+            use_agent = 2
+
+        if use_agent == 1:
+            port = self.get_num("Agent port: ", "50001", 5)
+            if port is False:
+                return False
+            elif port == -1:
+                self.close()
+
+            sock = socket(AF_INET, SOCK_STREAM)
+            sock.settimeout(60)
+            try:
+                sock.connect(('localhost', port))
+                sendmsg(sock, build_message((b'GET',)))
+            except OSError as err:
+                self.draw_text(False, (1, 0, err.__str__()),
+                                      (3, 0, "Press any key."))
+                if self.any_key() == -1:
+                    self.close()
+                return False
+
+            db_buf = receive(sock)
+            if db_buf[:4] == b'FAIL' or db_buf[:4] == b'[Err':
+                self.draw_text(False,
+                               (1, 0, db_buf),
+                               (3, 0, 'Press any key.'))
+                if self.any_key() == -1:
+                    self.close()
+                return False
+            sock.shutdown(SHUT_RDWR)
+            sock.close()
+
+            sock = socket(AF_INET, SOCK_STREAM)
+            sock.settimeout(60)
+            try:
+                sock.connect(('localhost', port))
+                sendmsg(sock, build_message((b'GETC',)))
+            except OSError as err:
+                self.draw_text(False, (1, 0, err.__str__()),
+                                      (3, 0, "Press any key."))
+                if self.any_key() == -1:
+                    self.close()
+                return False
+
+            answer = receive(sock)
+            parts = answer.split(b'\xB2\xEA\xC0')
+            password = parts.pop(0).decode()
+            keyfile_cont = parts.pop(0).decode()
+            if keyfile_cont == '':
+                keyfile = None
+            else:
+                if not isdir('/tmp/keepassc'):
+                    makedirs('/tmp/keepassc')
+                with open('/tmp/keepassc/tmp_keyfile', 'w') as handler:
+                    handler.write(parts.pop(0).decode())
+                    keyfile = '/tmp/keepassc/tmp_keyfile'
+
+            server = parts.pop(0).decode()
+            port = int(parts.pop(0))
+            if parts.pop(0) == b'True':
+                ssl = True
+            else:
+                ssl = False
+            tls_dir = parts.pop(0).decode()
+        elif use_agent is False:
+            return False
+        elif use_agent == -1:
+            self.close()
+        else:
+            if isfile(self.remote_home):
+                with open(self.remote_home, 'r') as handler:
+                    last_address = handler.readline()
+                    last_port = handler.readline()
+            else:
+                last_address = '127.0.0.1'
+                last_port = None
+
+            pass_auth = False
+            pass_ssl = False
+            while True:
+                if pass_auth is False:
+                    ret = self.get_authentication()
+                    if ret is False:
+                        return False
+                    elif ret == -1:
+                        self.close()
+                    password, keyfile = ret
+                pass_auth = True
+                if pass_ssl is False:
+                    ssl = self.gen_menu(1, ((1, 0, 'Use SSL/TLS (1)'),
+                                            (2, 0, 'Plain text (2)')))
+                    if ssl is False:
+                        pass_auth = False
+                        continue
+                    elif ssl == -1:
+                        self.close()
+                pass_ssl = True
+                server = Editor(self.stdscr, max_text_size=1,
+                                inittext=last_address,
+                                win_location=(0, 1),
+                                win_size=(1, self.xsize),
+                                title="Server address")()
+                if server is False:
+                    pass_ssl = False
+                    continue
+                elif server == -1:
+                    self.close()
+                if last_port is None:
+                    if ssl == 1:
+                        ssl = True # for later use
+                        std_port = "50003"
+                    else:
+                        ssl = False
+                        std_port = "50000"
+                else:
+                    if ssl == 1:
+                        ssl = True # for later use
+                    else:
+                        ssl = False
+                    std_port = last_port
+
+                port = self.get_num("Server port: ", std_port, 5)
+                if port is False:
+                    path_auth = True
+                    path_ssl = True
+                    continue
+                elif port == -1:
+                    self.close()
+                break
+
+            if ssl is True:
+                try:
+                    datapath = realpath(expanduser(getenv('XDG_DATA_HOME')))
+                except:
+                    datapath = realpath(expanduser('~/.local/share'))
+                finally:
+                    tls_dir = join(datapath, 'keepassc')
+            else:
+                tls_dir = None
+
+            client = Client(logging.INFO, 'client.log', server, port,
+                            password, keyfile, ssl, tls_dir)
+            db_buf = client.get_db()
+            if db_buf[:4] == 'FAIL' or db_buf[:4] == "[Err":
+                self.draw_text(False,
+                               (1, 0, db_buf),
+                               (3, 0, 'Press any key.'))
+                if self.any_key() == -1:
+                    self.close()
+                return False
+        self.db = KPDBv1(None, password, keyfile)
+        self.db.load(db_buf)
+        db = DBBrowser(self, True, server, port, ssl, tls_dir)
+        del db
+        return True
+
     def browser_help(self, mode_new):
         '''Print help for filebrowser'''
 
-        cur.nocbreak()
-        self.stdscr.keypad(0)
-        cur.endwin()
         if mode_new:
-            print('Navigate with arrow keys.')
-            print('\'o\' - choose directory')
-            print('\'e\' - abort')
-            print('\'H\' - show/hide hidden files')
-            print('\'ngg\' - move to line n')
-            print('\'G\' - move to last line')
-            print('/text - go to \'text\' (like in vim/ranger)')
+            self.draw_help(
+            'Navigate with arrow keys.',
+            '\'o\' - choose directory',
+            '\'e\' - abort',
+            '\'H\' - show/hide hidden files',
+            '\'ngg\' - move to line n',
+            '\'G\' - move to last line',
+            '/text - go to \'text\' (like in vim/ranger)',
+            '\n',
+            'Press return.')
         else:
-            print('Navigate with arrow keys.')
-            print('\'q\' - close program')
-            print('\'e\' - abort')
-            print('\'H\' - show/hide hidden files')
-            print('\'ngg\' - move to line n')
-            print('\'G\' - move to last line')
-            print('/text - go to \'text\' (like in vim/ranger)')
-        try:
-            input('Press return')
-        except EOFError:
-            print('')
-            if self.db is not None:
-                self.db.db_close()
-            exit()
-        self.initialize_cur()
+            self.draw_help(
+            'Navigate with arrow keys.',
+            '\'q\' - close program',
+            '\'e\' - abort',
+            '\'H\' - show/hide hidden files',
+            '\'ngg\' - move to line n',
+            '\'G\' - move to last line',
+            '/text - go to \'text\' (like in vim/ranger)',
+            '\n',
+            'Press return.')
 
     def dbbrowser_help(self):
-        cur.nocbreak()
-        self.stdscr.keypad(0)
-        cur.endwin()
-        print('\'e\' - go to main menu')
-        print('\'q\' - close program')
-        print('\'x\' - save db and close program')
-        print('\'s\' - save db')
-        print('\'S\' - save db with alternative filepath')
-        print('\'c\' - copy password of current entry')
-        print('\'b\' - copy username of current entry')
-        print('\'H\' - show password of current entry')
-        print('\'o\' - open URL of entry in standard webbrowser')
-        print('\'P\' - edit db password')
-        print('\'g\' - create group')
-        print('\'G\' - create subgroup')
-        print('\'y\' - create entry')
-        print('\'d\' - delete group or entry')
-        print('\'t\' - edit title of selected group or entry')
-        print('\'u\' - edit username')
-        print('\'p\' - edit password')
-        print('\'U\' - edit URL')
-        print('\'C\' - edit comment')
-        print('\'E\' - edit expiration date')
-        print('\'f\' or \'/\' - find entry by title')
-        print('\'L\' - lock db')
-        print('Navigate with arrow keys or h/j/k/l like in vim')
-        print('Type \'F5\' in a dialog to return to the previous one')
-        print('Type \'return\' to enter subgroups')
-        print('Type \'backspace\' to go back')
-        try:
-            input('Press return.')
-        except EOFError:
-            if not self.db is None:
-                self.db.db_close()
-            exit()
-        self.initialize_cur()
+        self.draw_help(
+        '\'e\' - go to main menu',
+        '\'q\' - close program',
+        '\'CTRL+D\' or \'CTRL+C\' - close program at any time',
+        '\'x\' - save db and close program',
+        '\'s\' - save db',
+        '\'S\' - save db with alternative filepath',
+        '\'c\' - copy password of current entry',
+        '\'b\' - copy username of current entry',
+        '\'H\' - show password of current entry',
+        '\'o\' - open URL of entry in standard webbrowser',
+        '\'P\' - edit db password',
+        '\'g\' - create group',
+        '\'G\' - create subgroup',
+        '\'y\' - create entry',
+        '\'d\' - delete group or entry (depends on what is marked)',
+        '\'t\' - edit title of selected group or entry',
+        '\'u\' - edit username',
+        '\'p\' - edit password',
+        '\'U\' - edit URL',
+        '\'C\' - edit comment',
+        '\'E\' - edit expiration date',
+        '\'f\' or \'/\' - find entry by title',
+        '\'L\' - lock db',
+        '\'m\' - enter move mode for marked group or entry',
+        '\'r\' - reload remote database (no function if not remote)',
+        'Navigate with arrow keys or h/j/k/l like in vim',
+        'Type \'return\' to enter subgroups',
+        'Type \'backspace\' to go back to parent',
+        'Type \'F5\' in a dialog to return to the previous one',
+        '\n',
+        'Press return.')
+
+    def move_help(self):
+        self.draw_help(
+        '\'e\' - go to main menu',
+        '\'q\' - close program',
+        '\'CTRL+D\' or \'CTRL+C\' - close program at any time',
+        'Navigate up or down with arrow keys or k and j',
+        'Navigate to subgroup with right arrow key or h',
+        'Navigate to parent with left arrow key or l',
+        'Type \'return\' to move the group to marked parent or the entry',
+        '\tto the marked group',
+        'Type \'backspace\' to move a group to the root',
+        'Type \'ESC\' to abort moving',
+        '\n',
+        'Press return.')
 
     def show_dir(self, highlight, dir_cont):
         '''List a directory with highlighting.'''
@@ -1036,7 +1299,8 @@ class Control(object):
 
         self.draw_text(changed)
         self.group_win.clear()
-        if parent is self.db._root_group:
+            
+        if parent is self.db.root_group:
             root_title = 'Parent: _ROOT_'
         else:
             root_title = 'Parent: ' + parent.title
@@ -1059,14 +1323,14 @@ class Control(object):
 
                 for i in range(num):
                     if highlight == i + offset:
-                        if groups[i].children:
+                        if groups[i + offset].children:
                             title = '+' + groups[i + offset].title
                         else:
                             title = ' ' + groups[i + offset].title
                         self.group_win.addstr(i + 1, 0, title,
                                               cur.color_pair(h_color))
                     else:
-                        if groups[i].children:
+                        if groups[i + offset].children:
                             title = '+' + groups[i + offset].title
                         else:
                             title = ' ' + groups[i + offset].title
@@ -1101,12 +1365,24 @@ class Control(object):
 
                 for i in range(num):
                     title = entries[i + offset].title
-                    if e_highlight == i + offset:
-                        self.entry_win.addstr(i, 2, title,
-                                              cur.color_pair(h_color))
+                    if date.today() > entries[i + offset].expire.date():
+                        expired = True
                     else:
-                        self.entry_win.addstr(i, 2, title,
-                                              cur.color_pair(n_color))
+                        expired = False
+                    if e_highlight == i + offset:
+                        if expired is True:
+                            self.entry_win.addstr(i, 2, title,
+                                                  cur.color_pair(3))
+                        else:
+                            self.entry_win.addstr(i, 2, title,
+                                                  cur.color_pair(h_color))
+                    else:
+                        if expired is True:
+                            self.entry_win.addstr(i, 2, title,
+                                                  cur.color_pair(7))
+                        else:
+                            self.entry_win.addstr(i, 2, title,
+                                                  cur.color_pair(n_color))
                 self.entry_win.addstr(ysize - 2, 2, (str(e_highlight + 1) +
                                                      ' of ' +
                                                      str(len(entries))))
@@ -1177,7 +1453,12 @@ class Control(object):
                 self.info_win.addstr(9, 0, expire)
                 if date.today() > entry.expire.date():
                     self.info_win.addstr(9, 22, ' (expired)')
-                self.info_win.addstr(10, 0, "Comment: " + comment)
+                if '\n' in comment:
+                    comment = comment.split('\n')[0]
+                    dots = ' ...'
+                else:
+                    dots = ''
+                self.info_win.addstr(10, 0, "Comment: " + comment + dots)
         except:
             pass
         finally:
