@@ -73,7 +73,10 @@ class Editor(object):
             except:
                 print('Invisible cursor not supported.')
         else:
-            curses.curs_set(1)
+            try:
+                curses.curs_set(1)
+            except:
+                pass
             curses.echo()
         locale.setlocale(locale.LC_ALL, '')
         curses.use_default_colors()
@@ -376,7 +379,7 @@ class Editor(object):
         return False
 
     def quit_nosave(self):
-        self.text = self.text_orig
+        self.text = False
         return False
 
     def help(self):
@@ -479,6 +482,8 @@ class Editor(object):
         curses.noecho()
         if self.text == -1:
             return -1
+        elif self.text is False:
+            return False
         else:
             return "\n".join(self.text)
 
